@@ -1,13 +1,15 @@
-import numpy.lib.stride_tricks
-import numpy as np
-from scipy.fft import rfft2 as rfft2_, irfft2 as irfft2_, fftshift as fftshift_
-from numpy import ma
-from scipy.signal import convolve2d as conv_
-from numpy import log
-
 """This module contains a pure python implementation of the basic
 cross-correlation algorithm for PIV image processing."""
 
+from typing import Optional, Tuple, Callable, Union
+import numpy.lib.stride_tricks
+import numpy as np
+from numpy import log
+from numpy import ma
+from numpy.fft import rfft2 as rfft2_, irfft2 as irfft2_, fftshift as fftshift_
+from scipy.signal import convolve2d as conv_
+
+
 __licence_ = """
 Copyright (C) 2011  www.openpiv.net
 
@@ -26,7 +28,11 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 """
 
 
-def get_field_shape(image_size, search_area_size, overlap):
+def get_field_shape(
+    image_size: Tuple[int,int],
+    search_area_size: Tuple[int,int],
+    overlap: Tuple[int,int],
+    )->Tuple[int,int]:
     """Compute the shape of the resulting flow field.
 
     Given the image size, the interrogation window size and
@@ -54,15 +60,22 @@ def get_field_shape(image_size, search_area_size, overlap):
     field_shape : 2-element tuple
         the shape of the resulting flow field
     """
-    field_shape = (np.array(image_size) - np.array(search_area_size)) // (
+    field_shape = (np.array(image_size) - np.array(search_area_size)) // (  # type: ignore
         np.array(search_area_size) - np.array(overlap)
     ) + 1
     
     return field_shape
 
 
-def get_coordinates(image_size, search_area_size, overlap, center_on_field = True):
+def get_coordinates(
+    image_size: Tuple[int,int],
+    search_area_size: int,
+    overlap: int,
+    center_on_field: bool=True
+    )->Tuple[np.ndarray, np.ndarray]:
     """Compute the x, y coordinates of the centers of the interrogation windows.
+    for the SQUARE windows only, see also get_rect_coordinates
+
     the origin (0,0) is like in the image, top left corner
     positive x is an increasing column index from left to right
     positive y is increasing row index, from top to bottom
@@ -100,10 +113,11 @@ def get_coordinates(image_size, search_area_size, overlap, center_on_field = Tru
 
     """
 
-    # get shape of the resulting flow field
+    # get shape of the resulting flow field as a 2 component array 
     field_shape = get_field_shape(image_size,
-                                  search_area_size,
-                                  overlap)
+                                  (search_area_size, search_area_size),
+                                  (overlap, overlap)
+                                  )
 
     # compute grid coordinates of the search area window centers
     # note the field_shape[1] (columns) for x
@@ -121,7 +135,7 @@ def get_coordinates(image_size, search_area_size, overlap, center_on_field = Tru
     # extreme left/right or top/bottom
     # have the same distance to the window edges. For simplicity only integer
     # movements are allowed.
-    if center_on_field == True:
+    if center_on_field is True:
         x += (
             image_size[1]
             - 1
@@ -137,24 +151,40 @@ def get_coordinates(image_size, search_area_size, overlap, center_on_field = Tru
         # the origin 0,0 is at top left
         # the units are pixels
 
-    return np.meshgrid(x, y)
+    X, Y = np.meshgrid(x,y)
+
+    return (X, Y)
 
 
-def get_rect_coordinates(frame_a, window_size, overlap, center_on_field = False):
+def get_rect_coordinates(
+    image_size: Tuple[int,int],
+    window_size: Union[int, Tuple[int,int]],
+    overlap: Union[int, Tuple[int,int]],
+    center_on_field: bool=False,
+    ):
     '''
     Rectangular grid version of get_coordinates.
     '''
-    if isinstance(window_size, tuple) == False and isinstance(window_size, list) == False:
-        window_size = [window_size, window_size]
-    if isinstance(overlap, tuple) == False and isinstance(overlap, list) == False:
-        overlap = [overlap, overlap]
-    _, y = get_coordinates(frame_a, window_size[0], overlap[0], center_on_field = False)
-    x, _ = get_coordinates(frame_a, window_size[1], overlap[1], center_on_field = False)
+    if isinstance(window_size, int):
+        window_size = (window_size, window_size)
+    if isinstance(overlap, int):
+        overlap = (overlap, overlap)
+
+    # @alexlib why the center_on_field is False? 
+    # todo: test True as well 
+    _, y = get_coordinates(image_size, window_size[0], overlap[0], center_on_field=center_on_field)
+    x, _ = get_coordinates(image_size, window_size[1], overlap[1], center_on_field=center_on_field)
+
+    X,Y = np.meshgrid(x[0,:], y[:,0])
     
-    return np.meshgrid(x[0,:], y[:,0])
+    return (X, Y)
 
 
-def sliding_window_array(image, window_size = 64, overlap = 32):
+def sliding_window_array(
+    image: np.ndarray, 
+    window_size: Tuple[int,int]=(64,64),
+    overlap: Tuple[int,int]=(32,32),
+    )-> np.ndarray:
     '''
     This version does not use numpy as_strided and is much more memory efficient.
     Basically, we have a 2d array and we want to perform cross-correlation
@@ -163,13 +193,14 @@ def sliding_window_array(image, window_size = 64, overlap = 32):
     with three dimension, of size (n_windows, window_size, window_size), in
     which each slice, (along the first axis) is an interrogation window. 
     '''
-    if isinstance(window_size, tuple) == False and isinstance(window_size, list) == False:
-        window_size = [window_size, window_size]
-    if isinstance(overlap, tuple) == False and isinstance(overlap, list) == False:
-        overlap = [overlap, overlap]
+    # if isinstance(window_size, int):
+    #     window_size = (window_size, window_size)
+    # if isinstance(overlap, int):
+    #     overlap = (overlap, overlap)
 
     x, y = get_rect_coordinates(image.shape, window_size, overlap, center_on_field = False)
-    x = (x - window_size[1]//2).astype(int); y = (y - window_size[0]//2).astype(int)
+    x = (x - window_size[1]//2).astype(int)
+    y = (y - window_size[0]//2).astype(int)
     x, y = np.reshape(x, (-1,1,1)), np.reshape(y, (-1,1,1))
 
     win_x, win_y = np.meshgrid(np.arange(0, window_size[1]), np.arange(0, window_size[0]))
@@ -446,7 +477,11 @@ def find_subpixel_peak_position(corr, subpixel_method="gaussian"):
         return subp_peak_position
 
 
-def sig2noise_ratio(correlation, sig2noise_method="peak2peak", width=2):
+def sig2noise_ratio(
+    correlation: np.ndarray,
+    sig2noise_method: str="peak2peak",
+    width: int=2
+    )-> np.ndarray:
     """
     Computes the signal to noise ratio from the correlation map.
 
@@ -501,13 +536,21 @@ def sig2noise_ratio(correlation, sig2noise_method="peak2peak", width=2):
                     corr, peak1_i, peak1_j, width=width
                 )
 
-                condition = (
-                    corr_max2 == 0
-                    or peak2_i == 0
+                border_condition = (
+                    peak2_i == 0
                     or peak2_i == corr.shape[0] - 1
                     or peak2_j == 0
                     or peak2_j == corr.shape[1] - 1
                 )
+
+                condition = (
+                    # non-valid second peak
+                    corr_max2 == 0
+                    or (
+                    # maybe a valid peak at the border, bad sign
+                    border_condition and corr_max2 > 0.5*corr_max1[i]
+                    )
+                )
                 if condition:  # mark failed peak2
                     corr_max2 = np.nan
 
@@ -600,7 +643,7 @@ def vectorized_sig2noise_ratio(correlation,
             out=np.zeros_like(peaks1),
             where=(peaks2 > 0.0)
         )
-        peak2peak[flag==True] = 0 # replace invalid values
+        peak2peak[flag is True] = 0 # replace invalid values
         return peak2peak
     
     elif sig2noise_method == "peak2mean":
@@ -621,19 +664,22 @@ def vectorized_sig2noise_ratio(correlation,
             out=np.zeros_like(peaks1max),
             where=(peaks2mean > 0.0)
         )
-        peak2mean[flag == True] = 0 # replace invalid values
+        peak2mean[flag is True] = 0 # replace invalid values
         return peak2mean
     else:
         raise ValueError(f"sig2noise_method not supported: {sig2noise_method}")
         
         
-def fft_correlate_images(image_a, image_b,
-                         correlation_method="circular",
-                         normalized_correlation=True,
-                         conj = np.conj,
-                         rfft2 = rfft2_,
-                         irfft2 = irfft2_,
-                         fftshift = fftshift_):
+def fft_correlate_images(
+    image_a: np.ndarray,
+    image_b: np.ndarray,
+    correlation_method: str="circular",
+    normalized_correlation: bool=True,
+    conj: Callable=np.conj,
+    rfft2 = rfft2_,
+    irfft2 = irfft2_,
+    fftshift = fftshift_,
+    )->np.ndarray:
     """ FFT based cross correlation
     of two images with multiple views of np.stride_tricks()
     The 2D FFT should be applied to the last two axes (-2,-1) and the
@@ -683,19 +729,19 @@ def fft_correlate_images(image_a, image_b,
     if correlation_method == "linear":
         # have to be normalized, mainly because of zero padding
         size = s1 + s2 - 1
-        fsize = 2 ** np.ceil(np.log2(size)).astype(int)
+        fsize = 2 ** np.ceil(np.log2(size)).astype(int) - 1
         fslice = (slice(0, image_a.shape[0]),
                   slice((fsize[0]-s1[0])//2, (fsize[0]+s1[0])//2),
                   slice((fsize[1]-s1[1])//2, (fsize[1]+s1[1])//2))
-        f2a = conj(rfft2(image_a, fsize, axes=(-2, -1)))
-        f2b = rfft2(image_b, fsize, axes=(-2, -1))
+        f2a = conj(rfft2(image_a, fsize, axes=(-2, -1)))  # type: ignore
+        f2b = rfft2(image_b, fsize, axes=(-2, -1))  # type: ignore
         corr = fftshift(irfft2(f2a * f2b).real, axes=(-2, -1))[fslice]
     elif correlation_method == "circular":
         f2a = conj(rfft2(image_a))
         f2b = rfft2(image_b)
         corr = fftshift(irfft2(f2a * f2b).real, axes=(-2, -1))
     else:
-        print("method is not implemented!")
+        print(f"correlation method {correlation_method } is not implemented")
 
     if normalized_correlation:
         corr = corr/(s2[0]*s2[1])  # for extended search area
@@ -793,7 +839,7 @@ def correlate_windows(window_a, window_b, correlation_method="fft",
     elif correlation_method == "direct":
         corr = convolve2d(window_a, window_b[::-1, ::-1], "full")
     else:
-        raise ValueError("method is not implemented")
+        print(f"correlation method {correlation_method } is not implemented")
 
     return corr
 
@@ -853,18 +899,18 @@ def fft_correlate_windows(window_a, window_b,
 
 
 def extended_search_area_piv(
-    frame_a,
-    frame_b,
-    window_size,
-    overlap=0,
-    dt=1.0,
-    search_area_size=None,
-    correlation_method="circular",
-    subpixel_method="gaussian",
-    sig2noise_method='peak2mean',
-    width=2,
-    normalized_correlation=False,
-    use_vectorized = False,
+    frame_a: np.ndarray,
+    frame_b: np.ndarray,
+    window_size: Union[int, Tuple[int,int]],
+    overlap: Union[int, Tuple[int,int]]=(0,0),
+    dt: float=1.0,
+    search_area_size: Optional[Union[int, Tuple[int,int]]]=None,
+    correlation_method: str="circular",
+    subpixel_method: str="gaussian",
+    sig2noise_method: Union[str, None]='peak2mean',
+    width: int=2,
+    normalized_correlation: bool=False,
+    use_vectorized: bool=False,
 ):
     """Standard PIV cross-correlation algorithm, with an option for
     extended area search that increased dynamic range. The search region
@@ -961,18 +1007,21 @@ def extended_search_area_piv(
     a NumPy vectorized solution in pyprocess.py
 
     """
-    if search_area_size is not None:
-        if isinstance(search_area_size, tuple) == False and isinstance(search_area_size, list) == False:
-            search_area_size = [search_area_size, search_area_size]
-    if isinstance(window_size, tuple) == False and isinstance(window_size, list) == False:
-        window_size = [window_size, window_size]
-    if isinstance(overlap, tuple) == False and isinstance(overlap, list) == False:
-        overlap = [overlap, overlap]
-        
-    # check the inputs for validity
+    # Reformat inputs so it works for both square and rectangular windows
+    # first if we get integer window size -> make it tuple
+    if isinstance(window_size, int):
+        window_size = (window_size, window_size)
+    # same for overlap
+    if isinstance(overlap, int):
+        overlap = (overlap, overlap)
+    
+    # if no search_size, copy window_size
     if search_area_size is None:
         search_area_size = window_size
+    elif isinstance(search_area_size, int):
+        search_area_size = (search_area_size, search_area_size)
 
+    # verify that things are logically possible: 
     if overlap[0] >= window_size[0] or overlap[1] >= window_size[1]:
         raise ValueError("Overlap has to be smaller than the window_size")
 
@@ -1014,7 +1063,7 @@ def extended_search_area_piv(
     corr = fft_correlate_images(aa, bb,
                                 correlation_method=correlation_method,
                                 normalized_correlation=normalized_correlation)
-    if use_vectorized == True:
+    if use_vectorized is True:
         u, v = vectorized_correlation_to_displacements(corr, n_rows, n_cols,
                                            subpixel_method=subpixel_method)
     else:
@@ -1023,7 +1072,7 @@ def extended_search_area_piv(
 
     # return output depending if user wanted sig2noise information
     if sig2noise_method is not None:
-        if use_vectorized == True:
+        if use_vectorized is True:
             sig2noise = vectorized_sig2noise_ratio(
                 corr, sig2noise_method=sig2noise_method, width=width
             )
@@ -1064,7 +1113,7 @@ def correlation_to_displacement(corr, n_rows, n_cols,
             # the find_subpixel_peak_position returns
             peak = np.array(find_subpixel_peak_position(corr[k*n_cols+m, :, :],
                             subpixel_method=subpixel_method)) -\
-                            default_peak_position
+                            default_peak_position  # type: ignore
 
         # the horizontal shift from left to right is the u
         # the vertical displacement from top to bottom (increasing row) is v
@@ -1075,11 +1124,11 @@ def correlation_to_displacement(corr, n_rows, n_cols,
     return (u, v)
 
 
-def vectorized_correlation_to_displacements(corr, 
-                                            n_rows = None,
-                                            n_cols = None,
-                                            subpixel_method = 'gaussian', 
-                                            eps = 1e-7
+def vectorized_correlation_to_displacements(corr: np.ndarray, 
+                                            n_rows: Optional[int]=None,
+                                            n_cols: Optional[int]=None,
+                                            subpixel_method: str='gaussian',
+                                            eps: float=1e-7
 ):
     """
     Correlation maps are converted to displacement for each interrogation
